import arc, os, pipes, re, posix, time
from datetime import datetime
from subprocess import Popen, PIPE
try:
    from subprocess import CalledProcessError
except ImportError:
    class CalledProcessError(OSError):
	def __init__(self, exitcode, cmd):
	    OSError.__init__(self, '%s exited with %d'%(cmd, exitcode))

terminal_job_states = \
    ['FINISHED', 'KILLED', 'FAILED', 'DELETED']
nonterminal_job_states = \
    ['ACCEPTING', 'ACCEPTED', 'EXECUTED', 'PREPARING',
     'SUBMITTING', 'SUBMITTED',
     'HOLD', 'QUEUING',
     'RUNNING', 'FINISHING']
all_job_states = terminal_job_states + nonterminal_job_states

class ParseError(Exception):
    pass

def explain_wait_status(status):
    if posix.WIFEXITED(status):
	exitcode = posix.WEXITSTATUS(status)
	if exitcode:
	    msg = 'exited with %d'%exitcode
	else:
	    msg = 'exited normally'
    elif posix.WIFSTOPPED(status):
	signo = posix.WSTOPSIG(status)
	msg = 'stopped by signal %d'%signo
    elif posix.WIFSIGNALED(status):
	signo = posix.WTERMSIG(status)
	msg = 'terminated by signal %d'%signo
    elif posix.WIFCONTINUED(status):
	msg = 'continued'
    if posix.WCOREDUMP(status):
	return '%s (core dumped)' % msg
    else:
	return msg

def _simple_cmd(argv, log = None):
    p = Popen(argv, stdout = PIPE, stderr = PIPE)
    output, errors = p.communicate()
    if log:
	if output:
	    for ln in output.strip().split('\n'):
		log.info('%s: %s'%(argv[0], ln))
	if errors:
	    for ln in errors.strip().split('\n'):
		log.error('%s: %s'%(argv[0], ln))
    if p.returncode != 0:
	raise CalledProcessError(p.returncode, ' '.join(argv))

def roundarg(t):
    return str(int(t + 0.5))

def arccp(src_url, dst_url, log = None, timeout = 20):
    return _simple_cmd(['arccp', '-t', roundarg(timeout), src_url, dst_url],
		       log = log)

def arcrm(dst_url, log = None, timeout = 20):
    return _simple_cmd(['arcrm', '-t', roundarg(timeout), dst_url], log = log)

def arcls(url, log = None, timeout = 20):
    p = Popen(['arcls', '-t', roundarg(timeout), '-l', url],
	      stdout = PIPE, stderr = PIPE)
    output, errors = p.communicate()
    entries = []
    for ln in output.split('\n')[1:]:
	if ln:
	    comps = ln.rsplit(None, 7)
	    entries.append(ArclsEntry(*comps))
    if p.returncode != 0:
	if log:
	    log.error('Errors from arcls:\n%s'%errors)
	raise CalledProcessError(p.returncode, 'arcls')
    return entries

def arcclean(jobids, force = False, timeout = 20, log = None):
    argv = ['arcclean', '-t', roundarg(timeout)]
    if force:
	argv.append('-f')
    argv += jobids
    _simple_cmd(argv, log = log)

_arcstat_state_re = re.compile(r'\w+\s+\(([^()]+)\)')
def parse_arcstat_state(s):
    if True:
	state = s.split(' ', 1)[0].upper()
    else:
	if s.startswith('Queuing '):
	    return 'QUEUING'
	if s.startswith('Running '):
	    return 'RUNNING'
	mo = re.match(_arcstat_state_re, s)
	if not mo:
	    raise ParseError('Malformed arcstat state %s.'%s)
	state = mo.group(1)
    if not state in all_job_states:
	raise ParseError('Unexpected job state %s'%state)
    return state

_arcstat_conv = {
    'State': parse_arcstat_state,
}

def arcstat(*jobids):
    cmd = 'arcstat ' + ' '.join(map(pipes.quote, jobids))
    fd = os.popen(cmd, 'r')
    jobstats = {}
    lnno = 0

    def parse_error(msg):
	fd.close()
	raise ParseError('Unexpected output from arcstat '
			 'at line %d: %s'%(lnno, msg))
    def check_entry(jobid, jobstat):
	if not 'State' in jobstat:
	    raise ParseError('Missing "State" key for %s.'%jobid)

    jobid, jobstat = None, None
    for ln in fd:
	ln = ln.strip()
	lnno += 1

	if ln == '':
	    if not jobid is None:
		check_entry(jobid, jobstat)
		jobstats[jobid] = jobstat
		jobid, jobstat = None, None
	elif ln == 'No jobs':
	    break
	elif ln.startswith('Job: '):
	    if not jobid is None:
		return parse_error('Missing empty line before "Job: ..."')
	    jobid = ln[5:].strip()
	    jobstat = {}
	elif ln.startswith('Warning:'):
	    pass
	else:
	    if jobid is None:
		return parse_error('Missing "Job: ..." header before %r'%ln)
	    try:
		k, v = ln.split(':', 1)
		if k in _arcstat_conv:
		    jobstat[k] = _arcstat_conv[k](v.strip())
		else:
		    jobstat[k] = v.strip()
	    except ValueError:
		return parse_error('Expecting "<key>: <value>", got %r'%ln)
    fd.close()
    if not jobid is None:
	check_entry(jobstat)
	jobstats[jobid] = jobstat
    return jobstats

def load_jobs_xml(timeout = 20, log = None):
    deadline = time.time() + timeout
    try:
	import xml.etree.ElementTree as et	# Python 2.6
    except ImportError:
	import elementtree.ElementTree as et	# Python 2.4
    uc = arc.UserConfig()
    jobstats = {}
    jp = os.path.join(uc.ARCUSERDIRECTORY, 'jobs.xml')
    # Should create a lock file, but better not make assumptions about ARC
    # internals.  This code should be oboselete when NG-3240 is closed.
    while os.path.exists(jp + '.lock'):
	if deadline - time.time() <= 1:
	    if log: log.warn('Timout while waiting for jobs.xml lock.')
	    return {}
	time.sleep(1)
    jx = et.parse(jp)
    for job in jx.findall('Job'):
	t_sub = job.findtext('LocalSubmissionTime')
	if t_sub:
	    t_sub = t_sub.replace('T', ' ')
	    if t_sub[-1] == 'Z':
		    t_sub = t_sub[0:-1]
	    jobstat = {'Submitted': t_sub}
	    jobstats[job.findtext('JobID')] = jobstat
    return jobstats

def arcprune(max_age = 604800, log = None, timeout = 20,
	     use_jobs_xml = False, force = True):
    t_start = time.time()
    pruned_count = 0
    failed_count = 0
    if use_jobs_xml:
	try:
	    jobstats = load_jobs_xml(timeout = 0.8*timeout, log = log)
	except ImportError, xc:
	    if log:
		log.warn('Missing dependency for custom jobs.xml loader: %s'%xc)
		log.warn('Falling back to arcstat.')
	    jobstats = arcstat('-al')
    else:
	jobstats = arcstat('-al')
    for jobid, jobstat in jobstats.iteritems():
	if not 'Submitted' in jobstat:
	    continue
	submitted = jobstat['Submitted']
	tm_sub = time.strptime(submitted, '%Y-%m-%d %H:%M:%S')
	t_sub = time.mktime(tm_sub)
	t_left = timeout - time.time() + t_start
	if t_left < 1:
	    if log: log.warn('Timout while pruning jobs.')
	    break
	if t_start - t_sub > max_age:
	    try:
		if log: log.info('Cleaning %s from %s.'%(jobid, submitted))
		arcclean([jobid], log = log, timeout = t_left, force = force)
		pruned_count += 1
	    except Exception, xc:
		failed_count = 0
		if log: log.warn('Failed to prune %s: %s'%(jobid, xc))
    return len(jobstats), pruned_count, failed_count

class ArclsEntry(object):
    DIR = 0
    FILE = 1

    def __init__(self, name, typ, size, cdate, ctod, validity, checksum, latency = ''):
	self.filename = name
	if typ == 'dir':
	    self.entry_type = self.DIR
	elif typ == 'file':
	    self.entry_type = self.FILE
	else:
	    self.entry_type = None
	self.size = size
#	Not in Python 2.4:
#	self.ctime = datetime.strptime(cdate + 'T' + ctod, '%Y-%m-%dT%H:%M:%S')
	def drop_NA(s):
	    if s != '(n/a)':
		return s

	if cdate == '(n/a)':
	    self.validity = drop_NA(validity)
	    self.checksum = drop_NA(checksum)
	    self.latency = drop_NA(latency)
	else:
	    self.validity = drop_NA(ctod)
	    self.checksum = drop_NA(validity)
	    self.latency = drop_NA(checksum)
