from arcnagios.nagutils import ServiceUNKNOWN
import ldap, os, random, re, time

def _subst_option(config, section, var, target_env):
    if config.has_option(section, 'default'):
	default = config.get(section, 'default', vars = target_env)
    else:
	default = None
    if default is None:
	raise ServiceUNKNOWN(
		'Missing required option (-O %s=...).'%var)
    target_env[var] = default

def _subst_getenv(config, section, var, target_env):
    if config.has_option(section, 'default'):
	default = config.get(section, 'default', vars = target_env)
    else:
	default = None
    if config.has_option(section, 'prefix'):
	prefix = config.get(section, 'prefix')
    else:
	prefix = ''
    if config.has_option(section, 'envvar'):
	v = os.getenv(config.get(section, 'envvar'), default)
    else:
	v = os.getenv(prefix + var, default)
    if v is None:
	raise ServiceUNKNOWN('Missing required option (-O) or '
			     'enviroment variable %s.'%var)
    target_env[var] = v

def _subst_ldap(config, section, var, target_env):
    lc = ldap.initialize(config.get(section, 'uri', vars = target_env))
    basedn = config.get(section, 'basedn', vars = target_env)
    filterstr = config.get(section, 'filter', vars = target_env)
    attr = config.get(section, 'attribute', vars = target_env)
    scope = ldap.SCOPE_SUBTREE
    try:
	sr = lc.search_s(basedn, scope, filterstr, attrlist = [attr])
    except ldap.LDAPError, xc:
	raise ServiceUNKNOWN('LDAP query %s failed with: %s' % (section, xc))
    for dn, ent in sr:
	if attr in ent:
	    target_env[var] = ent[attr][0]
	    return
    if config.has_option(section, 'default'):
	target_env[var] = config.get(section, 'default', vars = target_env)
    else:
	raise ServiceUNKNOWN('LDAP query %s did not provide a value for %s.'
			     % (filterstr, section))

def _subst_pipe(config, section, var, target_env):
    if not config.has_option(section, 'command'):
	raise ServiceUNKNOWN('Missing command in [variable.%s].'%var)
    fh = os.popen(config.get(section, 'command'))
    target_env[var] = fh.read().strip()
    fh.close()

def _subst_random_line(config, section, var, target_env):
    path = config.get(section, 'input_file', vars = target_env)
    try:
	fh = open(path)
	ln = fh.next().strip()
    except IOError, xc:
	raise ServiceUNKNOWN(str(xc))
    except StopIteration:
	raise ServiceUNKNOWN('%s must contains at least one line'%path)

    rng = random.Random(time.time())
    n = 1
    if config.has_option(section, 'exclude'):
	exclude = config.get(section, 'exclude', vars = target_env).split()
    else:
	exclude = []
    while True:
	try:
	    ln_next = fh.next().strip()
	    if not ln_next or ln_next.startswith('#') or ln_next in exclude:
		continue
	    if rng.randint(0, n) == 0:
		ln = ln_next
	except StopIteration:
	    fh.close()
	    target_env[var] = ln
	    return
	n += 1

def _subst_strftime(config, section, var, target_env):
    if config.has_option(section, 'raw_format'):
	format = config.get(section, 'raw_format', vars = target_env, raw=True)
    else:
	format = config.get(section, 'format', vars = target_env)
    target_env[var] = time.strftime(format)

def _subst_switch(config, section, var, target_env):
    case = 'case[%s]'%config.get(section, 'index', vars = target_env)
    if config.has_option(section, case):
	import_interpolated_variables(config, section, case, target_env)
	target_env[var] = config.get(section, case, vars = target_env)
    else:
	if not config.has_option(section, 'default'):
	    raise ServiceUNKNOWN(
		    'No %s and no default in section variable.%s.'%(case, var))
	import_interpolated_variables(config, section, 'default', target_env)
	target_env[var] = config.get(section, 'default', vars = target_env)

_method_by_name = {
    'getenv': _subst_getenv,
    'ldap': _subst_ldap,
    'option': _subst_option,
    'pipe': _subst_pipe,
    'random_line': _subst_random_line,
    'strftime': _subst_strftime,
    'switch': _subst_switch,
}

def register_substitution_method(name, f):
    _method_by_name[name] = f

def import_section_variables(config, section, target_env):
    """Import all variables listed in ``variables`` in ``section``."""

    if config.has_option(section, 'variables'):
	vars = config.get(section, 'variables').split()
	for var in vars:
	    if var in target_env:
		if target_env[var] == None:
		    raise ServiceUNKNOWN('Substitutions involving %s are '
					 'cyclic.'%var)
		continue
	    target_env[var] = None
	    import_variable(config, var, target_env)

_interp_re = re.compile(r'%\(([a-zA-Z0-9_]+)\)')

def import_interpolated_variables(config, section, var, target_env):
    """Import variables needed for expanding ``var`` in ``section``."""

    raw_value = config.get(section, var, raw = True)
    for mo in re.finditer(_interp_re, raw_value):
	v = mo.group(1)
	if not v in target_env:
	    import_variable(config, v, target_env)

def import_variable(config, var, target_env):
    """Import ``var`` by executing its defining section, populating
    ``target_env`` with its value and the values of any dependent
    variables."""

    section = 'variable.' + var
    import_section_variables(config, section, target_env)
    method = config.get(section, 'method')
    try:
	return _method_by_name[method](config, section, var, target_env)
    except KeyError:
	raise ServiceUNKNOWN('Unknown substitution method %s.'%method)
